import json

from mcclient.response import SLPResponse, LegacySLPResponse
from mcclient.base_client import BaseClient
from mcclient.encoding.packet import Packet


class SLPClient(BaseClient):
    def __init__(self, host, port=25565, timeout=5, proto=47, srv=True):
        super().__init__(host=host, port=port, timeout=timeout, proto=proto, srv=srv)
        self.retries = 0


    def _status_request(self):
        packet = Packet([b"\x00"]) # send status request
        packet = packet.pack()
        self._send(packet)
        res = self._recv()

        if res[0]: # if packetloss accured
            if self.retries < 3:
                self.retries += 1
                self._reset()
                return self._status_request()

            else:
                raise Exception("Max retries exceeded.")

        self._close(flush=False)

        res = res[2][2:]
        res = res.decode("utf-8")
        res = json.loads(res)
        return res


    def get_status(self):
        self._connect()
        self._handshake() # handshake + set connection state
        res = self._status_request()
        self.retries = 0
        return SLPResponse(self.hostname, self.port, res)



class LegacySLPClient(BaseClient):
    def __init__(self, host, port=25565, timeout=5):
        super().__init__(host=host, port=port, timeout=timeout)


    def get_status(self):
        self._connect()
        self._send(b"\xFE\x01") # legacy status request
        raw_res = self.sock.recv(1024)
        self._close()

        res = raw_res[3:] # remove padding and other headers
        res = res.decode("UTF-16-be", errors="ignore")
        res = res.split("\x00")
        res = LegacySLPResponse(self.hostname, self.port, res)
        return res


from mcclient.response import StatusResponse


def test_remove_color_codes():
    tests = [
        # basic tests
        ("§cHello §lworld§r!", False, "Hello world!"),
        ("§eHello §kworld§r!", True, "Hello world!"),
        ("Hello world!", False, "Hello world!"),

        # Tests with repeated color codes
        ("§c§cRed§r", False, "Red"),
        ("§e§eYellow§r", True, "Yellow"),

        # Tests with invalid color codes
        ("§xHello§y", False, "§xHello§y"),

        # Tests with different formatting codes
        ("§fHello §oWorld§r", False, "Hello World"),
        ("§fHello §oWorld§r", True, "Hello World"),

        # Tests with mixed formatting and color codes
        ("§e§lHello §c§oWorld§r", False, "Hello World"),
        ("§e§lHello §c§oWorld§r", True, "Hello World")
    ]

    test_res = StatusResponse("", 0, None)

    for cstr, bedrock, expected_output in tests:
        assert test_res._remove_color_codes(cstr, bedrock) == expected_output



from mcclient.encoding.packet import Packet
from mcclient.encoding.varint import VarInt



def test_varint_enc():
    # encoding
    enc_test_values = [1, 127, 128, 255]
    enc_test_results = [b"\x01", b"\x7f", b"\x80\x01", b"\xff\x01"]
    varint = VarInt()

    for value, result in zip(enc_test_values, enc_test_results):
        encoded_value = varint.pack(value)
        assert encoded_value == result


def test_packet_encoding():
    test_result = b"\x1A\x01\x15\x74\x68\x69\x73\x20\x69\x73\x20\x61\x20\x74\x65\x73\x74\x20\x73\x74\x72\x69\x6E\x67\x01\x00\x00"
    test_fields = (
        b"\x01",
        "this is a test string",
        "\x00",
        False
        )
    test_packet = Packet(test_fields)
    enc_test_packet = test_packet.pack()
    assert enc_test_packet == test_result


from mcclient.slp import SLPClient
from mcclient.slp import LegacySLPClient
from mcclient.base_client import BaseClient
from mcclient.bedrock_slp import BedrockSLPClient
from mcclient.query.query_client import QueryClient


import datetime
import re


class Players:
    def __init__(self, online, max, list):
        self.online = online
        self.max = max
        self.list = list


class Version:
    def __init__(self, name, protocol):
        self.name = name
        self.protocol = protocol


class StatusResponse:
    def __init__(self, host, port, raw_res):
        self.host = host
        self.port = port
        self.raw_res = raw_res
        self.res = {}

        self.res["host"] = self.host
        self.res["port"] = self.port
        self.timestamp = str(datetime.datetime.now())


    @staticmethod
    def _remove_color_codes(cstr: str, bedrock: bool = False) -> str:
        """
        Returns the input string stripped of all Minecraft color/formatting codes.
        documentation on minecraft color/formatting codes:
        https://minecraft.fandom.com/wiki/Formatting_codes

        Args:
        cstr (str): The string to remove color codes from.
        bedrock (bool): Wether to method should remove bedrock color/formating codes.

        Returns:
        str: The input string stripped of all Minecraft color/formatting codes.
        """

        if bedrock:
            return re.sub(r"§[a-u0-9]", "", cstr)
        return re.sub(r"§[a-fk-or0-9]", "", cstr)


class SLPResponse(StatusResponse):
    def __init__(self, host, port, raw_res):
        super().__init__(host, port, raw_res)

        self.res = self.res | self._parse_slp_res(self.raw_res)
        self.motd = self.res["motd"]
        self.favicon = self.res["favicon"]
        self.version = Version(
            self.res["version"]["name"], self.res["version"]["protocol"])
        self.players = Players(
            self.res["players"]["online"], self.res["players"]["max"], self.res["players"]["list"])

    def _parse_slp_res(self, slp_res):
        slp_res = self._add_missing(slp_res)

        if "sample" in slp_res["players"]:
            slp_res["players"]["list"] = slp_res["players"]["sample"]
            slp_res["players"].pop("sample")

        if "description" in slp_res:
            slp_res["motd"] = slp_res["description"]
            slp_res.pop("description")

        else:
            slp_res["motd"] = ""

        for player in slp_res["players"]["list"]:
            player["last_seen"] = self.timestamp

        if "favicon" in slp_res:
            slp_res["favicon"] = True

        else:
            slp_res["favicon"] = False

        slp_res["motd"] = self._parse_motd(slp_res["motd"])
        slp_res["version"]["name"] = self._remove_color_codes(slp_res["version"]["name"])
        slp_res["status"] = "online"
        return slp_res


    @classmethod
    def _parse_motd(cls, raw_motd):
        motd = ""
        if type(raw_motd) == dict:
            entries = raw_motd.get("extra", [])
            end = raw_motd.get("text", "")

            for entry in entries:
                motd += entry.get("text", "")
            motd += end

        elif type(raw_motd) == str:
            motd = raw_motd

        motd = motd.replace("\n", " ").strip()
        motd = cls._remove_color_codes(motd)
        return motd


    @staticmethod
    def _add_missing(res):
        default_res = {
            "previewsChat": None,
            "enforcesSecureChat": None,
            "description": "",
            "version": {"name": "", "protocol": -1},
            "players": {"online": -1, "max": -1, "list": []}
        }

        for key in default_res:
            if key not in res:
                res[key] = default_res[key]

        res["players"]["list"] = default_res["players"]["list"]
        return res



class LegacySLPResponse(StatusResponse):
    def __init__(self, host, port, raw_res):
        super().__init__(host, port, raw_res)

        self.res = self.res | self._parse_res(self.raw_res)
        self.motd = self.res["motd"]
        self.version = Version(self.res["version"], None)
        self.players = Players(self.res["online"], self.res["max"], None)


    @staticmethod
    def _parse_res(raw_res):
        res = {}
        res["version"] = raw_res[2]
        res["motd"] = raw_res[3]
        res["online"] = raw_res[4]
        res["max"] = raw_res[5]
        return res



class QueryResponse(StatusResponse):
    def __init__(self, host, port, raw_res):
        super().__init__(host, port, raw_res)

        self.res = raw_res | self.res

        self.motd = self.res["motd"]
        self.gametype = self.res["gametype"]
        self.game_id = self.res["game_id"]
        self.plugins = self.res["plugins"]
        self.map = self.res["map"]
        self.hostip = self.res["hostip"]
        self.hostport = self.res["hostport"]

        self.players = Players(
            self.res["numplayers"], self.res["maxplayers"], self.res["players"])
        self.version = Version(self.res["version"], None)
        self.version.software = self.res["software"]



class BedrockResponse(StatusResponse):
    def __init__(self, host, port, raw_res):
        super().__init__(host, port, raw_res)

        self.res = {}
        self.res["version"] = {}
        self.res["version"]["brand"] = self.raw_res[0]
        self.res["version"]["protocol"] = int(self.raw_res[2])
        self.res["version"]["name"] = self.raw_res[3]
        self.res["motd"] = self._remove_color_codes(self.raw_res[1])
        self.res["online_players"] = int(self.raw_res[4])
        self.res["max_players"] = int(self.raw_res[5])
        self.res["server_id"] = self.raw_res[6]

        self.res["map"] = None
        self.res["gametype"] = None
        if len(self.raw_res) > 6:
            self.res["map"] = self.raw_res[7]

        if len(self.raw_res) > 7:
            self.res["gametype"] = self.raw_res[8]

        self.version = Version(
            self.res["version"]["name"], self.res["version"]["protocol"])
        self.version.brand = self.res["version"]["brand"]
        self.motd = self.res["motd"]
        self.version = self.res["version"]
        self.online_players = self.res["online_players"]
        self.max_players = self.res["max_players"]
        self.server_id = self.res["server_id"]
        self.map = self.res["map"]
        self.gametype = self.res["gametype"]



import socket
import struct

from mcclient.address import Address
from mcclient.response import BedrockResponse


class BedrockSLPClient:
    def __init__(self, host, port=19132, timeout=4):
        self.get_host(host, port)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)


    def get_host(self, hostname, port):
        addr = Address(hostname, proto="udp")
        addr = addr.get_host()
        self.hostname = hostname
        self.host = addr[0]
        if addr[1] == None:
            self.port = port

        else:
            self.port = addr[1]


    def get_status(self):
        res = self._request_status()
        res = self._parse_res(res)
        return BedrockResponse(self.hostname, self.port, res)


    def _request_status(self): # needs an update (https://wiki.vg/Raknet_Protocol#Unconnected_Ping)
        status_request = b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x124Vx"
        self.sock.sendto(status_request, (self.host, self.port))
        return self.sock.recv(4096)


    @staticmethod
    def _parse_res(res):
        res = res[1:]
        extra_len = struct.unpack(">H", res[32:34])[0]
        res = res[34 : 34 + extra_len]
        res = res.decode()
        res = res.split(";")
        return res


import socket

from mcclient.address import Address
from mcclient.encoding.packet import Packet
from mcclient.encoding.varint import VarInt


class BaseClient:
    def __init__(self, host="localhost", port=25565, timeout=5, proto=47, srv=True):
        self.get_host(host, port, srv=srv)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(timeout)
        self.varint = VarInt()
        self.connected = False
        self.protocoll_version = self.varint.pack(proto)


    def get_host(self, hostname, port, srv):
        addr = Address(hostname)
        addr = addr.get_host(srv)
        self.hostname = hostname
        self.host = addr[0]
        if addr[1] == None:
            self.port = port

        else:
            self.port = addr[1]


    def _connect(self):
        if self.connected == False:
            self.sock.connect((self.host, self.port))
            self.connected = True

        elif self.connected == None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.connected = False
            self._connect()


    def _send(self, packet):
        return self.sock.send(packet)


    def _recv(self):
        length = self.varint.unpack(self.sock)
        packet_id = self.varint.unpack(self.sock)
        packet_id = self.varint.pack(packet_id)
        data = self.sock.recv(length)
        if len(data) < length - 4:
            loss = True
        
        else:
            loss = False
        return loss, packet_id, data


    def _close(self, flush=True):
        if flush:
            self._flush()
        self.sock.close()
        self.connected = None


    def _reset(self):
        self._close()
        self._connect()
        self._handshake()


    def _flush(self, length=8192):
        self.sock.recv(length)


    def implant_socket(self, sock):
        self.sock = sock
        self.connected = True


    def _handshake(self, next_state=1):
        fields = (
            b"\x00", # packet id
            self.protocoll_version,
            self.hostname,
            25565,
            self.varint.pack(next_state)# next state 1 for status request
        )
        packet = Packet(fields)
        packet = packet.pack()
        self._send(packet)


import struct


class VarInt: # class to pack and unpack Varints
    @staticmethod
    def pack(data):
        ordinal = b''
        while data != 0:
            byte = data & 0x7F
            data >>= 7
            ordinal += struct.pack('B', byte | (0x80 if data > 0 else 0))
        return ordinal


    @staticmethod
    def unpack(sock):
        data = 0
        for i in range(5):
            ordinal = sock.recv(1)
            if len(ordinal) == 0:
                break

            byte = ord(ordinal)
            data |= (byte & 0x7F) << 7*i
            if not byte & 0x80:
                break

        return data


import struct

from mcclient.encoding.varint import VarInt


class Packet:
    def __init__(self, fields=[]):
        self.fields = fields
        self.varint = VarInt()


    def pack(self):
        packet = b""
        for field in self.fields:
            field = self._encode(field)
            packet += field

        packet = self.varint.pack(len(packet)) + packet # add packetlength
        return packet


    def _encode(self, data):
        if type(data) == int:
            data = struct.pack(">H", data)

        elif type(data) == str:
            data = data.encode("utf-8")
            data = self.varint.pack(len(data)) + data

        elif type(data) == bool:
            data = b"\x01" if data else b"\x00"
        return data


from mcclient import BedrockSLPClient

from tests.utils import BaseTestConn, TooManyPackets


class BedrockTestConn(BaseTestConn):
    max_packets = 1
    def __init__(self):
        super().__init__()


    def sendto(self, data, *args):
        if self.packets == 0:
            self.packets += 1
            example_req = b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x124Vx"
            assert data == example_req

            example_res = b"\x1C\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\xF2\xA6\x29\x24\xC3\xD2\x00\xFF\xFF\x00\xFE\xFE\xFE\xFE\xFD\xFD\xFD\xFD\x12\x34\x56\x78\x00\x63\x4D\x43\x50\x45\x3B\xC2\xA7\x62\xC2\xA7\x6C\x45\x55\x20\xC2\xA7\x37\xC2\xA7\x6C\xC2\xBB\x20\x42\x4C\x4F\x43\x4B\x20\x50\x41\x52\x54\x59\x20\xEE\x84\x87\x3B\x31\x32\x31\x3B\x31\x2E\x30\x3B\x39\x39\x31\x32\x3B\x31\x30\x30\x30\x30\x31\x3B\x2D\x36\x30\x30\x31\x32\x38\x38\x31\x32\x30\x37\x30\x37\x37\x33\x31\x36\x37\x33\x3B\x48\x69\x76\x65\x20\x47\x61\x6D\x65\x73\x3B\x53\x75\x72\x76\x69\x76\x61\x6C"
            self.respond(example_res)

        elif self.packets >= self.max_packets:
            raise TooManyPackets(self.max_packets)



def test_bedrock():
    test_conn = BedrockTestConn()
    bedrock_client = BedrockSLPClient("example.com")
    bedrock_client.sock = test_conn # implant test socket
    bedrock_client.get_status()
    # Todo: add checks for response


import ipaddress

import dns.resolver


class Address:
    def __init__(self, addr, proto="tcp"):
        self.addr = addr
        self.proto = proto
        self.is_ip = self._ip_check(self.addr)


    def get_host(self, srv=True):
        if self.is_ip:
            return self.addr, None

        else:
            return self._resolve_hostname(self.addr, srv)


    def _resolve_hostname(self, hostname, srv):
        host = self._resolve_a_record(hostname)
        srv_record = None
        if srv:
            try:
                srv_record = self._mc_srv_lookup(hostname, self.proto)
                srv = True

            except Exception:
                pass

        if srv_record != None:
            return host, srv_record[1]

        else:
            return host, None


    @staticmethod
    def _mc_srv_lookup(hostname, proto):
        srv_prefix = f"_minecraft._{proto}."
        srv_record = dns.resolver.resolve(srv_prefix + hostname, "SRV")[0] # only use the first srv record returned
        host = str(srv_record.target).rstrip(".")
        port = int(srv_record.port)
        return host, port


    @staticmethod
    def _resolve_a_record(hostname):
        record = dns.resolver.resolve(hostname, "A")[0]
        return str(record).rstrip(".")


    @staticmethod
    def _ip_check(addr):
        try:
            ipaddress.ip_address(addr)
            return True
        except ValueError:
            return False


import json

from mcclient import SLPClient, LegacySLPClient
from mcclient.encoding.packet import Packet

from tests.utils import BaseTestConn, TooManyPackets

TEST_PROTO = 47
TEST_HOSTNAME = "example.com"
BASE_RES = {
        "description" : {
            "text": "This is an example motd"
        },
        "version": {
            "name": "1.19.2",
            "protocol": 795
        },
        "players": {
            "max": 69,
            "online": 5,
            "sample": [
                {
                "name": "thinkofdeath",
                "id": "4566e69f-c907-48ee-8d71-d7ba5aa00d20"
                }
            ]
        },
        "previewsChat": True,
        "enforcesSecureChat": True
    }


class SLPTestConn(BaseTestConn):
    max_packets = 2

    def __init__(self, hostname, base_res, proto):
        super().__init__()
        self.hostname = hostname
        self.base_res = base_res
        self.proto = proto


    def send(self, data):
        if self.packets == 0:
            self.packets += 1
            fields = (
                b"\x00", # packet id
                self.varint.pack(self.proto),
                self.hostname,
                25565,
                self.varint.pack(1)
            )
            test_packet = Packet(fields).pack()
            assert data == test_packet

        elif self.packets == 1:
            self.packets += 1
            assert data == b"\x01\x00"
            self.respond_res()

        elif self.packets >= self.max_packets:
            raise TooManyPackets(self.max_packets)


    def respond_res(self):
        res_fields = (
            b"\x00",
            json.dumps(self.base_res)
        )
        packet = Packet(res_fields).pack()
        self.respond(packet)



def test_slp_request():
    test_conn = SLPTestConn(TEST_HOSTNAME, BASE_RES, TEST_PROTO)
    slp_client = SLPClient(TEST_HOSTNAME, proto=TEST_PROTO)
    slp_client.implant_socket(test_conn)
    slp_client.get_status()



class LegaySLPTestConn(BaseTestConn):
    max_packets = 1

    def __init__(self, base_res):
        super().__init__()
        self.base_res = base_res


    def send(self, data):
        if self.packets == 0:
            self.packets += 1
            assert data == b"\xFE\x01"

            test_res = b"""\xFF\x00\x25\x00\xA7\x00\x31\x00\x00\x00
                \x31\x00\x32\x00\x37\x00\x00\x00\x31\x00\x2E\x00\x31
                \x00\x39\x00\x2E\x00\x33\x00\x00\x00\x41\x00\x20\x00
                \x4D\x00\x69\x00\x6E\x00\x65\x00\x63\x00\x72\x00\x61
                \x00\x66\x00\x74\x00\x20\x00\x53\x00\x65\x00\x72\x00
                \x76\x00\x65\x00\x72\x00\x00\x00\x30\x00\x00\x00\x32
                \x00\x30"""
            # test packet dump from raw leagy slp response, SHOULD be changed in the future!!
            self.respond(test_res)

        elif self.packets >= self.max_packets:
            raise TooManyPackets(self.max_packets)



def test_legacy_slp_request():
    test_conn = LegaySLPTestConn(BASE_RES)
    legacy_slp_client = LegacySLPClient(TEST_HOSTNAME)
    legacy_slp_client.implant_socket(test_conn)
    legacy_slp_client.get_status()


import socket
import struct
import random

from mcclient.response import QueryResponse
from mcclient.query.packet import QueryPacket


class QueryClient:
    def __init__(self, host, port=25565, timeout=0.5):
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)


    def _handshake(self):
        self.session_id = random.randint(0, 2147483648) & 0x0F0F0F0F # generate session id from int between 0 and 2147483648
        packet = QueryPacket(
            9, # type 9 for handshaking
            self.session_id,
            b"" # empty payload
        )
        packet = packet.pack()
        self._send(packet)
        res = self._recv()
        self.token = struct.pack('>l', int(res[2][:-1])) # extract token from response
        

    def _send(self, packet):
        return self.sock.sendto(packet, (self.host, self.port))


    def _recv(self):
        res = self.sock.recv(8192)
        type = res[0]
        session_id = res[1:5]
        payload = res[5:]
        return type, session_id, payload


    def _query_request(self):
        self._handshake()
        payload = self.token + b"\x00\x00\x00\x00" # challenge token and some padding for a full status request
        packet = QueryPacket(
            0, # packettype 0 for a status request
            self.session_id,
            payload
        )
        packet = packet.pack()
        self._send(packet)
        raw_res = self._recv()
        return self._read_query(raw_res)


    def get_status(self):
        res = self._query_request()
        return QueryResponse(self.host, self.port, res)


    @staticmethod
    def _read_query(res):
        res = res[2][11:] # remove unnecessary padding
        stats, players = res.split(b"\x00\x00\x01player_\x00\x00") # split stats from players
        data = {}

        stats = stats.split(b"\x00")
        stats = [stat.decode("utf-8") for stat in stats] # decode keys and values
        stats[0] = "motd" # replace "hostname" with "motd"
        key = True
        for y, x in enumerate(stats):
            if key:
                data[x] = stats[y + 1]
                key = False

            else:
                key = True

        for key in ["numplayers", "maxplayers", "hostport"]: # convert strings to ints
            data[key] = int(data[key])

        data["software"] = "Vanilla"
        software_parts = data["plugins"].split(":", 1)
        data["software"] = software_parts[0].strip()
        if len(software_parts) == 2:
            data["plugins"] = [plugin.strip() for plugin in software_parts[1].split(";")]

        else:
            data["plugins"] = []

        players = players[:-2] # remove endpadding
        players = players.split(b"\x00") # split players
        data["players"] = [player.decode("utf-8") for player in players if player != b""] # decode players
        return data


from mcclient.address import Address


def test_address():
    test_domain = Address("google.com").get_host()
    # add more tests for test_domain

    test_ip = Address("23.23.23.23").get_host()
    assert test_ip[0] == "23.23.23.23"

    test_srv_record = Address("pokecentral.org").get_host() # random server for testing, needs to be changed!
    assert test_srv_record[1] == 25565


import struct

class QueryPacket:
    def __init__(self, type, session_id, payload):
        self.type = type
        self.session_id = session_id
        self.payload = payload


    def pack(self):
        packet = b"\xFE\xFD" # some padding
        packet += struct.pack("!B", self.type)
        packet +=  struct.pack('>l', self.session_id)
        packet += self.payload
        return packet